/******************************************************************************
 *
 * scan.c - network scanner part of smap
 *
 * How to add a new test:
 *
 * - add new storage to fingerprint struct (i.e. fingerprint.newmethod)
 * - add symbol (i.e. SIP_NEWMETHOD) to config.h enumeration
 * - add the test itself to requestn.c:request_generate()
 * - invoke test by adding to scan_tests[] below
 * - add fingerprinting information to fingerprint.db
 *
 * $Id: scan.c,v 1.29 2007/05/10 13:45:31 hscholz Exp $
 *****************************************************************************/
#ifndef _SCAN_C
#define _SCAN_C

#include "config.h"

struct scan_test_t {
	int requesttype;
	int	offset;
	char name[256];
};
struct scan_test_t scan_tests[] = {
	{ SIP_OPTIONS, offsetof(fingerprint, options),
		"SIP OPTIONS request (valid)" },
	{ SIP_BROKENFROMTO, offsetof(fingerprint, brokenfromto),
		"SIP OPTIONS request with broken To/FROM URI (invalid)" }, 
	{ SIP_NEWMETHOD, offsetof(fingerprint, newmethod),
		"SIP request with unknown 'NEWMETHOD' request method (invalid)" },
	{ SIP_PRACK, offsetof(fingerprint, prack),
		"SIP PRACK to non-existent transaction" },
	{ SIP_INVITE, offsetof(fingerprint, invite),
		"SIP INVITE to non-existent extension" },
	{ SIP_PING, offsetof(fingerprint, ping),
		"SIP PING request (valid)" },
	{ SIP_TMP, offsetof(fingerprint, tmp), "scratchpad" },
	{ 0, 0, ""}
};

/******************************************************************************
 *
 * scan_host()
 *
 * test one host for SIP functionality
 * if fingerprinting is asked for we might do more tests than needed
 * if ICMP ping is asked for we do an ICMP ping before doing any SIP tests
 *
 *****************************************************************************/
int scan_host(struct in_addr dst) {

	extern struct config_t config;
	extern struct stats_t stats;

	int icmp, stoptests, testnum, sipenabled, *intp;
	struct scan_test_t test;

	char res[SIPLEN];

	fingerprint fp;

	FUNCTION(__FILE__, __FUNCTION__);

    stats.scanned++;

	if (FP_ENABLED)
		fingerprint_init(&fp);

	/* do ICMP ping if asked for and root priv. available */
	if (!(config.flags & FLAG_NOICMPPING)) {
		if (config.flags & FLAG_NOROOT) {
			icmp = 0;
		} else {
			icmp = icmpping(dst);
			if (icmp == 1) {
				stats.up++; /* host is up */
			} else {
				icmp = 0;
			}
		}
	} else {
		icmp = 0; /* assume ICMP failed, needed so we increase the	
				   * stats.up counter if SIP based response was received */
	}

	/* conduct all tests possible one by one
	 * In locating mode (fingerprinting disabled) we stop once we figured
	 * out the device speaks SIP
	 * In fingerprinting mode we do all tests and compare the fingerprint
	 * against the database
	 */

	stoptests = 0;	/* pre-maturely stop probing */
	testnum = 0;	/* first test */
	sipenabled = 0;	/* is host SIP-enabled? */
	test = (struct scan_test_t) scan_tests[testnum];
	while (!stoptests && (test.requesttype != 0)) {
		error(ERR_DEBUG, "running test: %s", test.name);

		bzero(&res, SIPLEN);
		/* use function pointers when adding TLS support */
		if (config.flags & FLAG_TCP_TRANSPORT) {
			if (!tcp_test(test.requesttype, dst, (char *) &res)) {
				intp = (void *) &(fp) + test.offset;
				*intp = NO_RESPONSE; /* no response indicator */
				goto nexttest;
			}
		} else {
			if (!udp_test(test.requesttype, dst, (char *) &res)) {
				intp = (void *) &(fp) + test.offset;
				*intp = NO_RESPONSE; /* no response indicator */
				goto nexttest;
			}
		}

		/* quick look at the response - is it SIP?
		 * it has to be >64Bytes and start with SIP/2.0
		 */
		if ((strlen(res) < SIPMINLEN) || strncmp("SIP/2.0 ", res, 8)) {
			goto nexttest;
		}
		if (!sipenabled) stats.sip++;	/* mark as SIP enabled for stats */
		sipenabled++;

		/* bail out by setting stoptests if fingerprinting wasn't
		 * asked for and device sent a SIP response */
		if (!(FP_ENABLED))
			stoptests++;

		/* fingerprinting was asked for, so do it */
		response_reason((void *) &(fp) + test.offset, res);

/*
		printf("response is: %d\n", (int) (&(fp) + test.offset));
		if (ack_needed((int) (&(fp) + test.offset))) {
			ack_send();
		}
*/
		
		
		/* look for some interesting headers to save for later output
		 * - User-Agent
		 * - Server
		 */
		if (!fp.header_ua)
			fp.header_ua = (char *) sip_getheader(res, "\r\nUser-Agent");
		if (!fp.header_server)
			fp.header_server = (char *) sip_getheader(res, "\r\nServer");

		/* check for headers */
		if ((fp.accept_class == NO_RELEVANCE)||(fp.accept_class == NO_RESPONSE))
			test_accept(&fp.accept_class, res);
		if ((fp.allow_class == NO_RELEVANCE)||(fp.allow_class == NO_RESPONSE))
			test_allow(&fp.allow_class, res);
		if (fp.supported_class == NO_RELEVANCE)
			test_supported(&fp.supported_class, res);
		/* we want the header order for options */
		if (test.requesttype == SIP_OPTIONS)
			test_headers(&fp.hoe_class, res);

nexttest:
		test = scan_tests[++testnum]; /* next test */
	}

	printf("Host %s:%d: (ICMP %s) SIP %s\n", inet_ntoa(dst),
		config.sip_port, icmp ? "OK" :
		((config.flags & FLAG_NOROOT) ? "untested" : "timeout" ),
		sipenabled? "enabled" : "timeout");
	if (!sipenabled) {
		error(ERR_DEBUG, "No answer received. Host down?");
	} else {
		if (FP_ENABLED) {
			/* this will generate the fingerprint output on stdout */
			fingerprint_lookup(fp);
		}
	}
	if (FP_ENABLED)
		fingerprint_free(fp);
	return (sipenabled? 1 : 0);
}

#endif /* _SCAN_C */
