/******************************************************************************
 *
 * transport_udp.c - UDP scanner part of smap
 *
 * $Id: transport_udp.c,v 1.4 2007/05/10 08:54:30 hscholz Exp $
 *****************************************************************************/
#ifndef _TRANSPORT_UDP_C
#define _TRANSPORT_UDP_C

#include "config.h"

/******************************************************************************
 *
 * udp_test()
 *
 * conduct a single SIP test and return the received response (might be
 * a SIP message) to the calling process
 * return 0 in case of error
 *
 * Socket-Mania:
 *
 * We have to use three ways to obtain all messages:
 * - stock UDP socket using sendto()
 * - RAW socket for ICMP error messages (if available)
 * - socket bound to local port to receive messages from broken clients
 *   (I didn't mention Cisco 79x0 phones ... did I?)
 * 
 *****************************************************************************/
int udp_test(unsigned int requesttype, struct in_addr dst, char *res) {

    int i;

    char req[SIPLEN];

    /* select() */
    fd_set fdset;
    struct timeval timeout;

	/* ssock = sendto() socket */
	/* bindsock = socket bound to local port */
	int ssock, bindsock, sockaddrlen;
#ifdef RAW_SOCKET
	int rawsock;
#endif
	struct sockaddr_in sockaddr, target;

	struct ip		*ippkt;
	struct icmp 	*icmppkt;
	char 			*errstr = NULL;

	unsigned int	l_callid, r_callid; /* unique Call-ID */
	int	endloop;	/* receive loop */

	FUNCTION(__FILE__, __FUNCTION__);

	bzero(&req, SIPLEN);

	/* sendto() socket */	
	ssock = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);
	if (ssock == -1) {
		if (DEBUG) perror("socket");
		goto error;
	}

#ifdef RAW_SOCKET
	/* try to obtain a raw socket for ICMP errors and UDP messages
	 * inbound from wrong source ports (hi Cisco!)
	 */
	rawsock = -1;
	if (!(config.flags & FLAG_NOROOT)) {
		rawsock = socket(PF_INET, SOCK_RAW, IPPROTO_ICMP);
		if (rawsock == -1) {
			error(ERR_NOTICE, "could not obtain RAW socket");
		} else {
			error(ERR_DEBUG, "RAW socket open");
		}
	}
#endif

	/* is 5060 (DEFAULT_SIP_PORT) available?
	 * if yes, we'll use it since some clients send responses to
	 * port 5060 even when instructed to not do so
	 */

	sockaddr.sin_family = PF_INET;
	sockaddr.sin_addr.s_addr = htonl(INADDR_ANY);
	sockaddr.sin_port = htons(DEFAULT_SIP_PORT);
	sockaddrlen = sizeof(sockaddr);

	bindsock = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);
	if (bindsock == -1) {
		/* socket() failed, default to random port */
		if (!(config.flags & FLAG_SIPPORTBUSY)) {
			error(ERR_NOTICE, "Could not obtain local port 5060."
				" Scanning may be unreliable!");
			config.flags |= FLAG_SIPPORTBUSY;
		}
		sockaddr.sin_family = PF_INET;
		sockaddr.sin_addr.s_addr = htonl(INADDR_ANY);
		sockaddr.sin_port = htons(0);
		sockaddrlen = sizeof(sockaddr);
	} else {
		/* try to bind() port 5060 */
		if (bind(bindsock, (struct sockaddr *) &sockaddr,
			sizeof(struct sockaddr)) == -1) {
			/* port 5060 in use */
			if (!(config.flags & FLAG_SIPPORTBUSY)) {
				error(ERR_NOTICE, "Could not obtain local port 5060."
					" Scanning may be unreliable!");
				config.flags |= FLAG_SIPPORTBUSY;
			}
			sockaddr.sin_family = PF_INET;
			sockaddr.sin_addr.s_addr = htonl(INADDR_ANY);
			sockaddr.sin_port = htons(0);
			sockaddrlen = sizeof(sockaddr);
		} else {
			/* success */
			sockaddr.sin_family = PF_INET;
			sockaddr.sin_addr.s_addr = htonl(INADDR_ANY);
			sockaddr.sin_port = htons(DEFAULT_SIP_PORT);
			sockaddrlen = sizeof(sockaddr);
		}
		close(bindsock);
	}

	/* bind() to local port in order to receive all packets
	 * do not bail out if error occurs
	 */
	bindsock = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);
	if (bindsock == -1) {
		error(ERR_DEBUG, "could not bind() to socket");
	} else {
		if (bind(bindsock, (struct sockaddr *) &sockaddr, sizeof(struct sockaddr)) == -1) {
			if (DEBUG) perror("bind");
			error(ERR_ERROR, "could not bind to port %d",
					ntohs(sockaddr.sin_port));
			goto error;
		} else {
			error(ERR_DEBUG, "bind() successful");
		}
	}

	if (getsockname(bindsock, (struct sockaddr *) &sockaddr,
				(socklen_t *) &sockaddrlen) != 0) 
	if (DEBUG) perror("getsocknamebindsock");

    if (request_generate(requesttype, (char *)&req, ntohs(sockaddr.sin_port), TRANSPORT_UDP, &l_callid) != 1) {
		fprintf(stderr, "\nERROR: SIP message assembly failed\n");
		goto error;
	}

    if (config.debug >= 2)
		printf("DEBUG: -- start request --\n%s"
			   "\nDEBUG: -- end of request --\n", req);

	target.sin_family = PF_INET;
    target.sin_addr = dst;
    target.sin_port = htons(config.sip_port);

    i = sendto(bindsock, req, strlen(req), 0, (struct sockaddr *) &target, sizeof(struct sockaddr));
	if (i != strlen(req)) {
		fprintf(stderr, "\nWarning: incomplete write()!\n");
		return 0;
	}


	/* write was ok, no we enter a receive loop which may in fact loop
	 * if we receive a retransmission of a previous request.
	 * we loop through this until we either find the response to our
	 * request or get caught in the timeout
	 */

	set_timeout(&timeout, config.timeout);
	do {

		/* prepare for select() */
		FD_ZERO(&fdset);
		FD_SET(ssock, &fdset);
		if (bindsock != -1)
			FD_SET(bindsock, &fdset);
#ifdef RAW_SOCKET
		if (rawsock != -1)
			FD_SET(rawsock, &fdset);
#endif

		if (select(FD_SETSIZE, &fdset, NULL, NULL, &timeout) > 0) {

			/* try all data sources */
			if (FD_ISSET(ssock, &fdset)) {
				i = read(ssock, res, SIPLEN);
			} else if (FD_ISSET(bindsock, &fdset)) {
				i = read(bindsock, res, SIPLEN);
#ifdef RAW_SOCKET
			} else if (FD_ISSET(rawsock, &fdset)) {
				i = read(rawsock, res, SIPLEN);
#endif
			}

			if (i <= 0) goto error;	

			/* check for icmp error message */
			ippkt = (struct ip *) res;
			if ((ippkt->ip_v == 4) && (ippkt->ip_p == IPPROTO_ICMP)) {
				/* ICMP */
				icmppkt = (struct icmp *) (res + (ippkt->ip_hl * 4));
				if ((errstr = icmp_message_str(icmppkt)) != NULL) { 
					error(ERR_DEBUG, "ICMP error %s", errstr);
					free(errstr);
				} else {
					error(ERR_DEBUG, "unknown ICMP message received (%d/%d)",
						icmppkt->icmp_type, icmppkt->icmp_code);
				}
				goto error;
			}
			/* DEBUG: print data */
			if (DEBUG && (i > 0)) {
				fprintf(stderr, 
				"DEBUG: -- start response --\n%.*s\n"
				"DEBUG: -- end of response --\n", i, res);
			}
			/* match unique Call-ID */
			r_callid = response_getcallid(res);
			if (r_callid == 0) {
				error(ERR_DEBUG, "could not parse Call-ID in received packet");
				/* r_callid might have been invalid but just to be sure
				 * we don't wait too long we bail out on this */
				endloop = 1;
			} else if (l_callid == r_callid) {
				endloop = 1;
			} else {
				/* probably received an old response, loop again */
				error(ERR_DEBUG, "ignoring possible SIP retransmission");
				endloop = 0;
			}
		} else {
			error(ERR_DEBUG, "select() timeout\n");
			goto error;
		}

	} while (endloop != 1); /* end of receive loop */

#ifdef RAW_SOCKET
	close(rawsock);
#endif
    close(ssock);
	if (bindsock != -1)
		close(bindsock);

	return 1;

error:
	if (DEBUG) printf("\n");
#ifdef RAW_SOCKET
	close(rawsock);
#endif
	close(ssock);
	if (bindsock != -1)
		close(bindsock);
	return 0;
}

#endif /* _TRANSPORT_UDP_C */
